-module(config).
-behaviour(gen_server).

-include_lib("mnesia_tables.hrl").

-export([status/0]).
-export([reload/0]).
-export([start_link/0, init/1, terminate/2]).
-export([handle_info/2, handle_cast/2, handle_call/3]).
-export([code_change/3]).

-record(state, {statuses}).

start_link() ->
    gen_server:start_link(?MODULE, [], []).

init([]) ->
    register(config, self()),
    Statuses = ets:new(log_statuses, []),
    %% TODO: reload?
    mnesia:activity(
      transaction,
      fun() ->
              mnesia:foldl(
                fun(#log_monitor_file{file = File, error_regex = ErrorRegex}, _Acc) ->
                        ets:insert(Statuses, {File, inactive}),
                        logfiles_sup:add_child([File, ErrorRegex]),
                        ok
                end, [], log_monitor_file)
      end),
    {ok, #state{statuses = Statuses}}.

handle_info({watcher_init, File}, State = #state{statuses = Statuses}) ->
    case ets:lookup(Statuses, File) of
        %% If the File has been removed, do nothing
        [] -> ok;
        _ -> ets:insert(Statuses, {File, active})
    end,
    {noreply, State};
handle_info({watcher_terminate, File}, State = #state{statuses = Statuses}) ->
    case ets:lookup(Statuses, File) of
        %% If the File has been removed, do nothing
        [] -> ok;
        _ -> ets:insert(Statuses, {File, inactive})
    end,
    {noreply, State};
handle_info(_Msg, State) ->
    {noreply, State}.

handle_cast(_Msg, State) ->
    {noreply, State}.

handle_call({reload}, _From, State = #state{statuses = Statuses}) ->
    {ok, File} = application:get_env(log_monitor, logfiles_config),
    {ok, Terms} = file:consult(File),
    MonitoredLogs = proplists:get_value(monitored_logs, Terms),
    Groups = [#log_monitor_group{name = Name, email_receivers = EmailReceivers} || {{group, Name}, {email_receivers, EmailReceivers}, _} <- MonitoredLogs],
    Logfiles = utils:flatten([[#log_monitor_file{file = F, error_regex = ErrorRegex, group = Name} || {{file, F}, {error_regex, ErrorRegex}} <- Logfiles]
                || {{group, Name}, {email_receivers, _}, {logfiles, Logfiles}} <- MonitoredLogs]),
    GroupsPartitions = compare_groups(Groups),
    manage_deleted_groups(proplists:get_value(deleted, GroupsPartitions)),
    manage_existing_groups(proplists:get_value(existing, GroupsPartitions)),
    manage_new_groups(proplists:get_value(new, GroupsPartitions)),
    LogfilesPartitions = compare_logfiles(Logfiles),
    manage_deleted_logfiles(proplists:get_value(deleted, LogfilesPartitions), Statuses),
    manage_existing_logfiles(proplists:get_value(existing, LogfilesPartitions)),
    manage_new_logfiles(proplists:get_value(new, LogfilesPartitions), Statuses),
    {reply, ok, State};
handle_call({get_statuses}, _From, State = #state{statuses = Statuses}) ->
    {reply, ets:tab2list(Statuses), State}.

terminate(_Reason, _State) ->
    shutdown.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

status() ->
    gen_server:call(config, {get_statuses}).

reload() ->
    gen_server:call(config, {reload}).

compare_groups(Groups) ->
    Old = list_groups(),
    Deleted = lists:filter(fun(Name) -> not lists:any(fun(#log_monitor_group{name = N, email_receivers = _}) -> N == Name end, Groups) end, Old),
    Existing = lists:filter(fun(#log_monitor_group{name = Name, email_receivers = _}) -> lists:member(Name, Old) end, Groups),
    New = lists:filter(fun(#log_monitor_group{name = Name, email_receivers = _}) -> not lists:member(Name, Old) end, Groups),
    [
     {deleted, Deleted},
     {existing, Existing},
     {new, New}
    ].

manage_deleted_groups(Groups) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(Name, _Acc) ->
                                  mnesia:delete({log_monitor_group, Name})
                          end, [], Groups)
      end).

manage_existing_groups(Groups) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(Group, _Acc) ->
                                  mnesia:write(Group)
                          end, [], Groups)
      end).

manage_new_groups(Groups) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(Group, _Acc) ->
                                  mnesia:write(Group)
                          end, [], Groups)
      end).

list_groups() ->
    mnesia:activity(
      transaction,
      fun() ->
              mnesia:foldl(
                fun(#log_monitor_group{name = Name, email_receivers = _}, Acc) ->
                        [Name | Acc]
                end, [], log_monitor_group)
      end).

compare_logfiles(Logfiles) ->
    Old= list_logfiles(),
    Deleted = lists:filter(fun(File) -> not lists:any(fun(#log_monitor_file{file = F, error_regex = _, group = _}) -> F == File end, Logfiles) end, Old),
    Existing = lists:filter(fun(#log_monitor_file{file = File, error_regex = _, group = _}) -> lists:member(File, Old) end, Logfiles),
    New = lists:filter(fun(#log_monitor_file{file = File, error_regex = _, group = _}) -> not lists:member(File, Old) end, Logfiles),
    [
     {deleted, Deleted},
     {existing, Existing},
     {new, New}
    ].

manage_deleted_logfiles(Logfiles, Statuses) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(File, _Acc) ->
                                  logfiles_sup:remove_child(File),
                                  mnesia:delete({log_monitor_file, File}),
                                  ets:delete(Statuses, File)
                          end, [], Logfiles)
      end).

manage_existing_logfiles(Logfiles) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(Logfile = #log_monitor_file{file = File, error_regex = ErrorRegex, group = _}, _Acc) ->
                                  mnesia:write(Logfile),
                                  case mnesia:read(log_monitor_file, File) of
                                      [#log_monitor_file{file = File, error_regex = ErrorRegex, group = _}] -> ok;
                                      _ ->
                                          %% The error regex has changed
                                          logfiles_sup:remove_child(File),
                                          logfiles_sup:add_child([File, ErrorRegex])
                                  end
                          end, [], Logfiles)
      end).

manage_new_logfiles(Logfiles, Statuses) ->
    mnesia:activity(
      transaction,
      fun() ->
              lists:foldl(fun(Logfile = #log_monitor_file{file = File, error_regex = ErrorRegex, group = _}, _Acc) ->
                                  mnesia:write(Logfile),
                                  ets:insert(Statuses, {File, inactive}),
                                  logfiles_sup:add_child([File, ErrorRegex])
                          end, [], Logfiles)
      end).

list_logfiles() ->
    mnesia:activity(
      transaction,
      fun() ->
              mnesia:foldl(
                fun(#log_monitor_file{file = File, error_regex = _, group = _}, Acc) ->
                        [File | Acc]
                end, [], log_monitor_file)
      end).
