-module(mailer).
-behaviour(gen_server).

-include_lib("mnesia_tables.hrl").

-export([start_link/0, init/1, terminate/2]).
-export([handle_info/2, handle_cast/2, handle_call/3]).
-export([code_change/3]).

start_link() ->
    gen_server:start_link(?MODULE, [], []).

init([]) ->
    register(mailer, self()),
    timer:send_after(1000, {send_emails}),
    {ok, []}.

handle_info({error, File, Text}, State) ->
    mnesia:activity(
      transaction,
      fun() ->
              Id = erlang:monotonic_time(),
              mnesia:write(#log_monitor_error{id = Id, file = File, text = Text})
      end),
    {noreply, State};
handle_info({send_emails}, State) ->
    Emails = emails_to_send(),
    try send_emails(Emails) of
        _ -> timer:send_after(1000, {send_emails})
    catch
        _Throw ->
            error_logger:info_msg("Waiting 60s before sending new emails~n"),
            timer:send_after(60000, {send_emails})
    end,
    {noreply, State}.

handle_cast(_Msg, State) ->
    {noreply, State}.

handle_call(_Request, _From, State) ->
    {noreply, State}.

terminate(_Reason, _State) ->
    shutdown.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

emails_to_send() ->
    mnesia:activity(
      transaction,
      fun() ->
              mnesia:foldl(
                fun(#log_monitor_error{id = Id, file = File, text = Text}, Acc) ->
                        [{Id, File, Text}] ++ Acc
                end, [], log_monitor_error)
      end).

send_emails(Emails) ->
    case Emails of
        [] -> {ok};
        [{Id, File, Text} | Others] ->
            case send_email(File, Text) of
                {error, Reason, Message} ->
                    error_logger:error_msg("Error sending email: ~s ~p~n", [Reason, Message]),
                    throw(cannot_send_email);
                _ ->
                    remove_email(Id),
                    send_emails(Others),
                    {ok}
            end
    end.

send_email(File, Text) ->
    {ok, EmailConfig} = application:get_env(log_monitor, email_config),
    Sender = proplists:get_value(sender, EmailConfig),
    DefaultReceiver = proplists:get_value(default_receiver, EmailConfig),
    RawSubject = proplists:get_value(subject, EmailConfig),
    Connection = proplists:get_value(connection, EmailConfig),
    try receivers(File) of
        GroupReceivers ->
            Receivers = if GroupReceivers == []
                           -> [DefaultReceiver];
                           true -> GroupReceivers
                        end,
            Subject = email_subject(RawSubject, File),
            gen_smtp_client:send_blocking(
              {Sender, Receivers,
               io_lib:format("Subject: ~s\r\nFrom: ~s\r\n\r\nLogfile: ~s~nError: ~n~s~n", [Subject, Sender, File, Text])},
              Connection
             )
    catch
        error:_ ->
            ok
    end.

remove_email(Id) ->
    mnesia:activity(
      transaction,
      fun() ->
              mnesia:delete({log_monitor_error, Id})
      end),
    ok.

receivers(File) ->
    mnesia:activity(
      transaction,
      fun() ->
              case mnesia:read(log_monitor_file, File) of
                  [{log_monitor_file, File, _, Group}] ->
                      case mnesia:read(log_monitor_group, Group) of
                          [{log_monitor_group, Group, EmailReceivers}] ->
                              EmailReceivers;
                          _ -> throw(file_group_not_found)
                      end;
                  _ -> throw(file_not_found)
              end
      end).

email_subject(Text, File) ->
    Text1 = re:replace(Text, "%f", File, [global, {return, list}]),
    re:replace(Text1, "%F", file_name_from_path(File), [global, {return, list}]).

file_name_from_path(File) ->
    {match, [_, Basename]} = re:run(File, "^.*\/(.*)\\..*$", [{capture, all, list}]),
    Basename.
